from collections.abc import Iterable
from typing import Optional, cast
from unittest import TestCase
from unittest.mock import Mock, call, patch, sentinel

from click.testing import CliRunner
from epplib.commands import Login, Logout

from eppic.commands.epp_command import EppCommand, LoginCommand
from eppic.models import ArgumentField
from eppic.settings import InteractiveSettings, Session

from .utils import DEFAULT_SESSIONS, TestClient, TestSettings, get_app


class LogoutCommandTest(TestCase):
    def test_cli(self):
        app = get_app()
        mock_client = cast(TestClient, app.epp_handler._client).mock

        app.run_command("logout")

        self.assertEqual(mock_client.mock_calls, [call(Logout(), tr_id=None), call.close()])
        self.assertEqual(app.epp_handler.session, sentinel.session)
        self.assertFalse(app.epp_handler.logged_in)

    def test_cli_cltrid(self):
        app = get_app()
        mock_client = cast(TestClient, app.epp_handler._client).mock

        app.run_command("logout --cltrid future-echo")

        self.assertEqual(mock_client.mock_calls, [call(Logout(), tr_id="future-echo"), call.close()])
        self.assertEqual(app.epp_handler.session, sentinel.session)
        self.assertFalse(app.epp_handler.logged_in)

    def test_confirm(self):
        data = [
            ("n", []),
            ("y", [call(Logout(), tr_id=None), call.close()]),
        ]
        for answer, calls in data:
            with self.subTest(answer=answer):
                app = get_app(TestSettings(interactive=InteractiveSettings(command_confirmation=True)))
                mock_client = cast(TestClient, app.epp_handler._client).mock
                with CliRunner().isolation(input=answer):
                    app.run_command("logout")

                    self.assertEqual(mock_client.mock_calls, calls)


@patch("eppic.epp.SocketTransport")
class LoginCommandTest(TestCase):
    default_obj_uris = Session.model_fields["obj_uris"].default

    @staticmethod
    def get_app(*args, **kwargs):
        """Return a test application with an already configured EPP session."""
        app = get_app(*args, **kwargs)
        app.epp_handler.session = DEFAULT_SESSIONS[0]
        return app

    def command(
        self, extensions: Optional[Iterable[ArgumentField]] = None, command_class: type[Login] = Login
    ) -> EppCommand:
        return LoginCommand("foobar", command_class, Mock(), extensions=extensions)

    def test_required_args(self, socket_transport):
        app = self.get_app()

        with patch("eppic.epp.Client") as mock_client:
            mock_client.return_value.send.return_value.code = 1000
            app.run_command("epp-login --cl-id=qwert --password=fdsa")

        self.assertTrue(app.epp_handler.logged_in)
        self.assertEqual(
            app.epp_handler.session,
            DEFAULT_SESSIONS[0].model_copy(
                update={"username": "qwert", "password": "fdsa", "obj_uris": self.default_obj_uris}
            ),
        )
        self.assertEqual(
            mock_client.mock_calls,
            [
                call(socket_transport.return_value),
                call().connect(),
                call().send(Login("qwert", "fdsa", self.default_obj_uris), tr_id=None),
            ],
        )

    def test_obj_uris(self, socket_transport):
        app = self.get_app()

        with patch("eppic.epp.Client") as mock_client:
            mock_client.return_value.send.return_value.code = 1000
            app.run_command("epp-login --cl-id=qwert --password=fdsa --obj-uris-1=ghjkl")

        self.assertTrue(app.epp_handler.logged_in)
        self.assertEqual(
            app.epp_handler.session,
            DEFAULT_SESSIONS[0].model_copy(update={"username": "qwert", "password": "fdsa", "obj_uris": ["ghjkl"]}),
        )
        self.assertEqual(
            mock_client.mock_calls,
            [
                call(socket_transport.return_value),
                call().connect(),
                call().send(Login("qwert", "fdsa", obj_uris=["ghjkl"]), tr_id=None),
            ],
        )

    def test_cltrid(self, socket_transport):
        app = self.get_app()

        with patch("eppic.epp.Client") as mock_client:
            mock_client.return_value.send.return_value.code = 1000
            app.run_command("epp-login --cl-id=qwert --password=fdsa --cltrid future-echo")

        self.assertTrue(app.epp_handler.logged_in)
        self.assertEqual(
            app.epp_handler.session,
            DEFAULT_SESSIONS[0].model_copy(
                update={"username": "qwert", "password": "fdsa", "obj_uris": self.default_obj_uris}
            ),
        )
        self.assertEqual(
            mock_client.mock_calls,
            [
                call(socket_transport.return_value),
                call().connect(),
                call().send(Login("qwert", "fdsa", self.default_obj_uris), tr_id="future-echo"),
            ],
        )

    def test_generate(self, socket_transport):
        # command class, extensions, input args, result command, result extensions
        cases: list[tuple[type[Login], list, dict, Login, list]] = [
            # no extensions
            (Login, [], {"cl_id": "foo", "password": "bar", "obj_uris": ["asdf"]}, Login("foo", "bar", ["asdf"]), []),
            # extensions exist, but not used
            (
                Login,
                [ArgumentField("foobar", str, False, sentinel.default)],
                {"cl_id": "foo", "password": "bar", "obj_uris": ["asdf"]},
                Login("foo", "bar", ["asdf"]),
                [],
            ),
            # extensions exist and used
            (
                Login,
                [ArgumentField("foobar", str, False, sentinel.default)],
                {"cl_id": "foo", "password": "bar", "foobar": "asdf", "obj_uris": ["asdf"]},
                Login("foo", "bar", ["asdf"]),
                ["asdf"],
            ),
            # cltrid set
            (
                Login,
                [ArgumentField("foobar", str, False, sentinel.default)],
                {"cl_id": "foo", "password": "bar", "foobar": "asdf", "obj_uris": ["asdf"], "cltrid": "foobarbaz"},
                Login("foo", "bar", ["asdf"]),
                ["asdf"],
            ),
            (
                Login,
                [ArgumentField("foobar", str, False, sentinel.default)],
                {"cl_id": "foo", "password": "bar", "obj_uris": ["asdf"], "cltrid": "asdf"},
                Login("foo", "bar", ["asdf"]),
                [],
            ),
            (
                Login,
                [],
                {"cl_id": "foo", "password": "bar", "obj_uris": ["asdf"], "cltrid": "asdf"},
                Login("foo", "bar", ["asdf"]),
                [],
            ),
        ]

        for cls, extensions, args, expected, expected_ext in cases:
            with self.subTest(cls=cls, extensions=extensions, args=args):
                cmd = self.command(extensions=extensions, command_class=cls)
                res = cmd._prepare_request(args)
                self.assertEqual(res, expected)
                self.assertEqual(cast(Login, res).extensions, expected_ext)

    def test_confirm(self, socket_transport):
        data = [
            ("n", []),
            (
                "y",
                [
                    call(socket_transport.return_value),
                    call().connect(),
                    call().send(Login("qwert", "fdsa", self.default_obj_uris), tr_id=None),
                ],
            ),
        ]
        for answer, calls in data:
            with self.subTest(answer=answer):
                app = self.get_app(TestSettings(interactive=InteractiveSettings(command_confirmation=True)))
                with CliRunner().isolation(input=answer):
                    with patch("eppic.epp.Client") as mock_client:
                        mock_client.return_value.send.return_value.code = 1000
                        app.run_command("epp-login --cl-id=qwert --password=fdsa")

                    self.assertEqual(mock_client.mock_calls, calls)
